@@ -705,7 +705,7 @@ int mode;
 	    errno = EACCES;
 	  UserReturn(errno);
 	}
-      if (q = UserStatus())
+      if ((q = UserStatus()))
 	{
 	  if (q > 0)
 	    errno = q;
