@@ -178,7 +178,7 @@ InitLoadav()
   debug("Init Kmem...\n");
   if ((kmemf = open("/dev/kmem", O_RDONLY)) == -1)
     return;
-# ifndef _AUX_SOURCE
+# if !defined(_AUX_SOURCE) && !defined(AUX)
 #  ifdef NLIST_NAME_UNION
   nl[0].n_un.n_name = LOADAV_AVENRUN;
 #  else
