@@ -21,7 +21,7 @@
  */
 
 #include "rcs.h"
-RCS_ID("$Id: termcap.c,v 1.3 1993/07/21 15:43:35 mlschroe Exp $ FAU")
+RCS_ID("$Id: termcap.c,v 1.8 1994/05/31 12:33:13 mlschroe Exp $ FAU")
 
 #include <sys/types.h>
 #include "config.h"
@@ -36,11 +36,26 @@ static char *findcap __P((char *, char **, int));
 static char *e_tgetstr __P((char *, char **));
 static int   e_tgetflag __P((char *));
 static int   e_tgetnum __P((char *));
+#ifdef MAPKEYS
+static int   addmapseq __P((char *, int));
+static int   remmapseq __P((char *));
+#ifdef DEBUG
+static void dumpmap __P((void));
+#endif
+#endif
+
 
 extern struct term term[];	/* terminal capabilities */
 extern struct NewWindow nwin_undef, nwin_default, nwin_options;
 extern int force_vt, assume_LP;
 extern int Z0width, Z1width;
+#ifdef MAPKEYS
+extern struct action umtab[];
+extern struct action mmtab[];
+extern struct action dmtab[];
+extern char *kmap_extras[];
+extern int kmap_extras_fl[];
+#endif
 
 char Termcap[TERMCAP_BUFSIZE + 8];	/* new termcap +8:"TERMCAP=" */
 static int Termcaplen;
@@ -69,7 +84,7 @@ char *s;
       if (term[i].type != T_STR)
 	continue;
       if (strcmp(term[i].tcname, s) == 0)
-	return d_tcs[i].str;
+	return D_tcs[i].str;
     }
   return 0;
 }
@@ -82,13 +97,14 @@ int he;
   register char *s;
   int i;
   char tbuf[TERMCAP_BUFSIZE], *tp;
+  int t, xue, xse, xme;
 
   ASSERT(display);
   bzero(tbuf, sizeof(tbuf));
-  debug1("InitTermcap: looking for tgetent('%s')\n", d_termname);
-  if (tgetent(tbuf, d_termname) != 1)
+  debug1("InitTermcap: looking for tgetent('%s')\n", D_termname);
+  if (*D_termname == 0 || tgetent(tbuf, D_termname) != 1)
     {
-      Msg(0, "Cannot find termcap entry for %s.", d_termname);
+      Msg(0, "Cannot find termcap entry for '%s'.", D_termname);
       return -1;
     }
   debug1("got it:\n%s\n",tbuf);
@@ -98,212 +114,461 @@ int he;
   if (extra_outcap)
     debug1("Extra outcap: %s\n", extra_outcap);
 #endif
-  tp = d_tentry;
+  tp = D_tentry;
 
   for (i = 0; i < T_N; i++)
     {
       switch(term[i].type)
 	{
 	case T_FLG:
-	  d_tcs[i].flg = e_tgetflag(term[i].tcname);
+	  D_tcs[i].flg = e_tgetflag(term[i].tcname);
 	  break;
 	case T_NUM:
-	  d_tcs[i].num = e_tgetnum(term[i].tcname);
+	  D_tcs[i].num = e_tgetnum(term[i].tcname);
 	  break;
 	case T_STR:
-	  d_tcs[i].str = e_tgetstr(term[i].tcname, &tp);
+	  D_tcs[i].str = e_tgetstr(term[i].tcname, &tp);
 	  /* no empty strings, please */
-	  if (d_tcs[i].str && *d_tcs[i].str == 0)
-	    d_tcs[i].str = 0;
+	  if (D_tcs[i].str && *D_tcs[i].str == 0)
+	    D_tcs[i].str = 0;
 	  break;
 	default:
 	  Panic(0, "Illegal tc type in entry #%d", i);
 	  /*NOTREACHED*/
 	}
     }
-  if (HC)
+  if (D_HC)
     {
       Msg(0, "You can't run screen on a hardcopy terminal.");
       return -1;
     }
-  if (OS)
+  if (D_OS)
     {
       Msg(0, "You can't run screen on a terminal that overstrikes.");
       return -1;
     }
-  if (NS)
-    {
-      Msg(0, "Terminal must support scrolling.");
-      return -1;
-    }
-  if (!CL)
+  if (!D_CL)
     {
       Msg(0, "Clear screen capability required.");
       return -1;
     }
-  if (!CM)
+  if (!D_CM)
     {
       Msg(0, "Addressable cursor capability required.");
       return -1;
     }
   if ((s = getenv("COLUMNS")) && (i = atoi(s)) > 0)
-    CO = i;
+    D_CO = i;
   if ((s = getenv("LINES")) && (i = atoi(s)) > 0)
-    LI = i;
+    D_LI = i;
   if (wi)
-    CO = wi;
+    D_CO = wi;
   if (he)
-    LI = he;
-  if (CO <= 0)
-    CO = 80;
-  if (LI <= 0)
-    LI = 24;
+    D_LI = he;
+  if (D_CO <= 0)
+    D_CO = 80;
+  if (D_LI <= 0)
+    D_LI = 24;
 
   if (nwin_options.flowflag == nwin_undef.flowflag)
-    nwin_default.flowflag = CNF ? FLOW_NOW * 0 : 
-			    XO ? FLOW_NOW * 1 :
+    nwin_default.flowflag = D_CNF ? FLOW_NOW * 0 : 
+			    D_NX ? FLOW_NOW * 1 :
 			    FLOW_AUTOFLAG;
-  CLP |= assume_LP || !AM || XV || XN ||
-	 (!extra_incap && !strncmp(d_termname, "vt", 2));
-  if (!(BL = e_tgetstr("bl", &tp)))
-  if (!BL)
-    BL = "\007";
-  if (!BC)
+  D_CLP |= assume_LP || !D_AM || D_XV || D_XN ||
+	 (!extra_incap && !strncmp(D_termname, "vt", 2));
+  if (!D_BL)
+    D_BL = "\007";
+  if (!D_BC)
     {
-      if (BS)
-	BC = "\b";
+      if (D_BS)
+	D_BC = "\b";
       else
-	BC = LE;
+	D_BC = D_LE;
     }
-  if (!CR)
-    CR = "\r";
-  if (!NL)
-    NL = "\n";
-  if (SG <= 0 && UG <= 0)
-    {
-      /*
-       * Does ME also reverse the effect of SO and/or US?  This is not
-       * clearly specified by the termcap manual. Anyway, we should at
-       * least look whether ME and SE/UE are equal:
-       */
-      if (UE && ((SE && strcmp(SE, UE) == 0) || (ME && strcmp(ME, UE) == 0)))
-	UE = 0;
-      if (SE && (ME && strcmp(ME, SE) == 0))
-	SE = 0;
+  if (!D_CR)
+    D_CR = "\r";
+  if (!D_NL)
+    D_NL = "\n";
 
-      for (i = 0; i < NATTR; i++)
-	d_attrtab[i] = d_tcs[T_ATTR + i].str;
-      /* Set up missing entries */
-      s = 0;
-      for (i = NATTR-1; i >= 0; i--)
-	if (d_attrtab[i])
-	  s = d_attrtab[i];
-      for (i = 0; i < NATTR; i++)
+  /*
+   *  Set up attribute handling.
+   *  This is rather complicated because termcap has different
+   *  attribute groups.
+   */
+
+  if (D_UG > 0)
+    D_US = D_UE = 0;
+  if (D_SG > 0)
+    D_SO = D_SE = 0;
+  /* Unfortunatelly there is no 'mg' capability.
+   * For now we think that mg > 0 if sg and ug > 0.
+   */
+  if (D_UG > 0 && D_SG > 0)
+    D_MH = D_MD = D_MR = D_MB = D_ME = 0;
+
+  xue = ATYP_U;
+  xse = ATYP_S;
+  xme = ATYP_M;
+
+  if (D_SO && D_SE == 0)
+    {
+      Msg(0, "Warning: 'so' but no 'se' capability.");
+      if (D_ME)
+	xse = xme;
+      else
+	D_SO = 0;
+    }
+  if (D_US && D_UE == 0)
+    {
+      Msg(0, "Warning: 'us' but no 'ue' capability.");
+      if (D_ME)
+	xue = xme;
+      else
+	D_US = 0;
+    }
+  if ((D_MH || D_MD || D_MR || D_MB) && D_ME == 0)
+    {
+      Msg(0, "Warning: 'm?' but no 'me' capability.");
+      D_MH = D_MD = D_MR = D_MB = 0;
+    }
+  /*
+   * Does ME also reverse the effect of SO and/or US?  This is not
+   * clearly specified by the termcap manual. Anyway, we should at
+   * least look whether ME and SE/UE are equal:
+   */
+  if (D_UE && D_SE && strcmp(D_SE, D_UE) == 0)
+    xse = xue;
+  if (D_SE && D_ME && strcmp(D_ME, D_SE) == 0)
+    xse = xme;
+  if (D_UE && D_ME && strcmp(D_ME, D_UE) == 0)
+    xue = xme;
+
+  for (i = 0; i < NATTR; i++)
+    {
+      D_attrtab[i] = D_tcs[T_ATTR + i].str;
+      D_attrtyp[i] = i == ATTR_SO ? xse : (i == ATTR_US ? xue : xme);
+    }
+  
+  /* Set up missing entries (attributes are priority ordered) */
+  s = 0;
+  t = 0;
+  for (i = 0; i < NATTR; i++)
+    if ((s = D_attrtab[i]))
+      {
+	t = D_attrtyp[i];
+	break;
+      }
+  for (i = 0; i < NATTR; i++)
+    {
+      if (D_attrtab[i] == 0)
 	{
-	  if (d_attrtab[i] == 0)
-	    d_attrtab[i] = s;
-	  else
-	    s = d_attrtab[i];
+	  D_attrtab[i] = s;
+	  D_attrtyp[i] = t;
 	}
+      else
+        {
+	  s = D_attrtab[i];
+	  t = D_attrtyp[i];
+        }
     }
-  else
-    {
-      MS = 1;
-      for (i = 0; i < NATTR; i++)
-	d_attrtab[i] = d_tcs[T_ATTR + i].str = 0;
-    }
-  if (!DO)
-    DO = NL;
-  if (!SF)
-    SF = NL;
-  if (IN)
-    IC = IM = 0;
-  if (EI == 0)
-    IM = 0;
+
+  if (!D_DO)
+    D_DO = D_NL;
+  if (!D_SF)
+    D_SF = D_NL;
+  if (D_IN)
+    D_IC = D_IM = 0;
+  if (D_EI == 0)
+    D_IM = 0;
   /* some strange termcap entries have IC == IM */
-  if (IC && IM && strcmp(IC, IM) == 0)
-    IC = 0;
-  if (KE == 0)
-    KS = 0;
-  if (CCE == 0)
-    CCS = 0;
-  if (CG0)
+  if (D_IC && D_IM && strcmp(D_IC, D_IM) == 0)
+    D_IC = 0;
+  if (D_KE == 0)
+    D_KS = 0;
+  if (D_CVN == 0)
+    D_CVR = 0;
+  if (D_VE == 0)
+    D_VI = 0;
+  if (D_CCE == 0)
+    D_CCS = 0;
+  if (D_CG0)
     {
-      if (CS0 == 0)
+      if (D_CS0 == 0)
 #ifdef TERMINFO
-        CS0 = "\033(%p1%c";
+        D_CS0 = "\033(%p1%c";
 #else
-        CS0 = "\033(%.";
+        D_CS0 = "\033(%.";
 #endif
-      if (CE0 == 0)
-        CE0 = "\033(B";
+      if (D_CE0 == 0)
+        D_CE0 = "\033(B";
+      D_AC = 0;
     }
-  else if (AS && AE)
+  else if (D_AC || (D_AS && D_AE))	/* some kind of graphics */
     {
-      CG0 = 1;
-      CS0 = AS;
-      CE0 = AE;
-      CC0 = AC;
+      D_CG0 = 1;
+      D_CS0 = (D_AS && D_AE) ? D_AS : "";
+      D_CE0 = (D_AS && D_AE) ? D_AE : "";
+      D_CC0 = D_AC;
     }
   else
     {
-      CS0 = CE0 = "";
-      CC0 = "g.h.i'j-k-l-m-n+o~p\"q-r-s_t+u+v+w+x|y<z>";
+      D_CS0 = D_CE0 = "";
+      D_CC0 = 0;
+      D_AC = "";	/* enable default string */
     }
+
   for (i = 0; i < 256; i++)
-    d_c0_tab[i] = i;
-  if (CC0)
-    for (i = strlen(CC0) & ~1; i >= 0; i -= 2)
-      d_c0_tab[(int)(unsigned char)CC0[i]] = CC0[i + 1];
-  debug1("ISO2022 = %d\n", CG0);
-  if (PF == 0)
-    PO = 0;
-  debug2("terminal size is %d, %d (says TERMCAP)\n", CO, LI);
+    D_c0_tab[i] = i;
+  if (D_AC)
+    {
+      /* init with default string first */
+      s = "l+m+k+j+u+t+v+w+q-x|n+o~s_p\"r#`+a:f'g#~o.v-^+<,>h#I#0#y<z>";
+      for (i = strlen(s) & ~1; i >= 0; i -= 2)
+	D_c0_tab[(int)(unsigned char)s[i]] = s[i + 1];
+    }
+  if (D_CC0)
+    for (i = strlen(D_CC0) & ~1; i >= 0; i -= 2)
+      D_c0_tab[(int)(unsigned char)D_CC0[i]] = D_CC0[i + 1];
+  debug1("ISO2022 = %d\n", D_CG0);
+  if (D_PF == 0)
+    D_PO = 0;
+  debug2("terminal size is %d, %d (says TERMCAP)\n", D_CO, D_LI);
 
   /* Termcap fields Z0 & Z1 contain width-changing sequences. */
-  if (CZ1 == 0)
-    CZ0 = 0;
+  if (D_CZ1 == 0)
+    D_CZ0 = 0;
   Z0width = 132;
   Z1width = 80;
 
   CheckScreenSize(0);
 
-  if (TS == 0 || FS == 0 || DS == 0)
-    HS = 0;
-  if (HS)
+  if (D_TS == 0 || D_FS == 0 || D_DS == 0)
+    D_HS = 0;
+  if (D_HS)
     {
       debug("oy! we have a hardware status line, says termcap\n");
-      if (WS <= 0)
-        WS = d_width;
+      if (D_WS <= 0)
+        D_WS = D_width;
     }
-
-  d_UPcost = CalcCost(UP);
-  d_DOcost = CalcCost(DO);
-  d_NLcost = CalcCost(NL);
-  d_LEcost = CalcCost(BC);
-  d_NDcost = CalcCost(ND);
-  d_CRcost = CalcCost(CR);
-  d_IMcost = CalcCost(IM);
-  d_EIcost = CalcCost(EI);
-
-#ifdef AUTO_NUKE
-  if (CAN)
+#ifdef KANJI
+  D_kanji = 0;
+  if (D_CKJ)
     {
-      debug("termcap has AN, setting autonuke\n");
-      d_auto_nuke = 1;
+      if (strcmp(D_CKJ, "euc") == 0)
+	D_kanji = EUC;
+      else if (strcmp(D_CKJ, "sjis") == 0)
+	D_kanji = SJIS;
     }
 #endif
-  if (COL > 0)
-    {
-      debug1("termcap has OL (%d), setting limit\n", COL);
-      d_obufmax = COL;
-    }
 
-  d_tcinited = 1;
+  D_UPcost = CalcCost(D_UP);
+  D_DOcost = CalcCost(D_DO);
+  D_NLcost = CalcCost(D_NL);
+  D_LEcost = CalcCost(D_BC);
+  D_NDcost = CalcCost(D_ND);
+  D_CRcost = CalcCost(D_CR);
+  D_IMcost = CalcCost(D_IM);
+  D_EIcost = CalcCost(D_EI);
+
+#ifdef AUTO_NUKE
+  if (D_CAN)
+    {
+      debug("termcap has AN, setting autonuke\n");
+      D_auto_nuke = 1;
+    }
+#endif
+  if (D_COL > 0)
+    {
+      debug1("termcap has OL (%d), setting limit\n", D_COL);
+      D_obufmax = D_COL;
+    }
+#ifdef MAPKEYS
+  D_nseqs = 0;
+  for (i = 0; i < T_OCAPS - T_CAPS; i++)
+    remap(i, 1);
+  for (i = 0; i < KMAP_EXT; i++)
+    remap(i + (KMAP_KEYS+KMAP_AKEYS), 1);
+  D_seqp = D_kmaps[0].seq;
+#endif
+
+  D_tcinited = 1;
   MakeTermcap(0);
   return 0;
 }
 
+#ifdef MAPKEYS
+
+int
+remap(n, map)
+int n;
+int map;
+{
+  char *s;
+  int fl = 0, domap = 0;
+  struct action *a1, *a2, *tab;
+
+  tab = umtab;
+  for (;;)
+    {
+      a1 = &tab[n];
+      a2 = 0;
+      if (n < KMAP_KEYS+KMAP_AKEYS)
+	{
+	  if (n >= KMAP_KEYS)
+	    n -= T_OCAPS-T_CURSOR;
+	  s = D_tcs[n + T_CAPS].str;
+	  if (n >= T_CURSOR-T_CAPS)
+	    a2 = &tab[n + (T_OCAPS-T_CURSOR)];
+	}
+      else
+	{
+	  s = kmap_extras[n - (KMAP_KEYS+KMAP_AKEYS)];
+	  fl |= kmap_extras_fl[n - (KMAP_KEYS+KMAP_AKEYS)];
+	}
+      if (s == 0)
+	return 0;
+      if (a1 && a1->nr == RC_ILLEGAL)
+	a1 = 0;
+      if (a2 && a2->nr == RC_ILLEGAL)
+	a2 = 0;
+      if (a1 && a1->nr == RC_STUFF && strcmp(a1->args[0], s) == 0)
+	a1 = 0;
+      if (a2 && a2->nr == RC_STUFF && strcmp(a2->args[0], s) == 0)
+	a2 = 0;
+      domap |= a1 || a2;
+      if (tab == umtab)
+	tab = dmtab;
+      else if (tab == dmtab)
+	tab = mmtab;
+      else
+	break;
+    }
+
+  if (map == 0 && domap)
+    return 0;
+  if (map && !domap)
+    return 0;
+  debug3("%smapping %s %#x\n", map? "" :"un",s,n);
+  if (map)
+    return addmapseq(s, n | fl);
+  else
+    return remmapseq(s);
+}
+
+static int
+addmapseq(seq, nr)
+char *seq;
+int nr;
+{
+  int i, j = 0, k, mo, m;
+  char *p, *o;
+
+  k = strlen(seq);
+  if (k > sizeof(D_kmaps[0].seq) - 1)
+    return -1;
+  for (i = 0; i < D_nseqs; i++)
+    if ((j = strcmp(D_kmaps[i].seq, seq)) >= 0)
+      break;
+  if (i < D_nseqs && j == 0)
+    {
+      D_kmaps[i].nr = nr;
+      return 0;
+    }
+  if (D_nseqs >= sizeof(D_kmaps)/sizeof(*D_kmaps)) /* just in case... */
+    return -1;
+  for (j = D_nseqs - 1; j >= i; j--)
+    D_kmaps[j + 1] = D_kmaps[j];
+  p = D_kmaps[i].seq;
+  o = D_kmaps[i].off;
+  strcpy(p, seq);
+  bzero(o, k + 1);
+  D_kmaps[i].nr = nr;
+  D_nseqs++;
+
+  if (i + 1 < D_nseqs)
+    for (j = 0; *p; p++, o++, j++)
+      {
+        if (D_kmaps[i + 1].seq[j] != *p)
+          {
+            if (D_kmaps[i + 1].seq[j])
+	      *o = 1;
+	    break;
+	  }
+        *o = D_kmaps[i + 1].off[j] ? D_kmaps[i + 1].off[j] + 1 : 0;
+      }
+
+  for (k = 0; k < i; k++)
+    for (m = j = 0, p = D_kmaps[k].seq, o = D_kmaps[k].off; *p; p++, o++, j++)
+      {
+	mo = m;
+	if (!m && *p != D_kmaps[i].seq[j])
+	  m = 1;
+	if (*o == 0 && mo == 0 && m)
+	  *o = i - k;
+	if (*o < i - k || (*o == i - k && m))
+	  continue;
+	(*o)++;
+      }
+#ifdef DEBUG
+  dumpmap();
+#endif
+  return 0;
+}
+
+static int
+remmapseq(seq)
+char *seq;
+{
+  int i, j = 0, k;
+  char *p, *o;
+
+  for (i = 0; i < D_nseqs; i++)
+    if ((j = strcmp(D_kmaps[i].seq, seq)) >= 0)
+      break;
+  if (i == D_nseqs || j)
+    return -1;
+  for (k = 0; k < i; k++)
+    for (j = 0, p = D_kmaps[k].seq, o = D_kmaps[k].off; *p; p++, o++, j++)
+      {
+	if (k + *o == i)
+          *o = D_kmaps[i].off[j] ? D_kmaps[i].off[j] + *o - 1 : 0;
+	else if (k + *o > i)
+          (*o)--;
+      }
+  for (j = i + 1; j < D_nseqs; j++)
+    D_kmaps[j - 1] = D_kmaps[j];
+  D_nseqs--;
+#ifdef DEBUG
+  dumpmap();
+#endif
+  return 0;
+}
+
+#ifdef DEBUG
+static void
+dumpmap()
+{
+  char *p, *o;
+  int i,j,n;
+  debug("Mappings:\n");
+  for (i = 0; i < D_nseqs; i++)
+    {
+      for (j = 0, p = D_kmaps[i].seq, o = D_kmaps[i].off; *p; p++, o++, j++)
+	{
+	  if (*p > ' ' && (unsigned char)*p < 0177)
+	    {
+              debug2("%c[%d] ", *p, *o);
+	    }
+          else
+            debug2("\\%03o[%d] ", (unsigned char)*p, *o);
+	}
+      n = D_kmaps[i].nr;
+      debug2(" ==> %d%s\n", n & ~KMAP_NOTIMEOUT, (n & KMAP_NOTIMEOUT) ? " (no timeout)" : "");
+    }
+}
+#endif
+
+#endif
 
 static void
 AddCap(s)
@@ -311,13 +576,13 @@ char *s;
 {
   register int n;
 
-  if (tcLineLen + (n = strlen(s)) > 55 && Termcaplen < TERMCAP_BUFSIZE - 4)
+  if (tcLineLen + (n = strlen(s)) > 55 && Termcaplen < TERMCAP_BUFSIZE + 8 - 4)
     {
       strcpy(Termcap + Termcaplen, "\\\n\t:");
       Termcaplen += 4;
       tcLineLen = 0;
     }
-  if (Termcaplen + n < TERMCAP_BUFSIZE)
+  if (Termcaplen + n < TERMCAP_BUFSIZE + 8)
     {
       strcpy(Termcap + Termcaplen, s);
       Termcaplen += n;
@@ -337,9 +602,9 @@ int aflag;
 
   if (display)
     {
-      wi = d_width;
-      he = d_height;
-      tname = d_termname;
+      wi = D_width;
+      he = D_height;
+      tname = D_termname;
     }
   else
     {
@@ -350,7 +615,7 @@ int aflag;
   debug1("MakeTermcap(%d)\n", aflag);
   if ((s = getenv("SCREENCAP")) && strlen(s) < TERMCAP_BUFSIZE)
     {
-      sprintf(Termcap, "TERMCAP=%s", s);     /* TERMCAP_BUFSIZE + ... ? XXX */
+      sprintf(Termcap, "TERMCAP=%s", s);
       sprintf(Term, "TERM=screen");
       debug("getenvSCREENCAP o.k.\n");
       return Termcap;
@@ -387,7 +652,7 @@ int aflag;
   while (0);		/* Goto free programming... */
   tcLineLen = 100;	/* Force NL */
   sprintf(Termcap,
-	  "TERMCAP=SC|%s|VT 100/ANSI X3.64 virtual terminal|", Term + 5);
+	  "TERMCAP=SC|%s|VT 100/ANSI X3.64 virtual terminal", Term + 5);
   Termcaplen = strlen(Termcap);
   debug1("MakeTermcap decided '%s'\n", p);
   if (extra_outcap && *extra_outcap)
@@ -410,110 +675,149 @@ int aflag;
   sprintf(buf, "li#%d:co#%d:", he, wi);
   AddCap(buf);
   AddCap("am:");
-  if (aflag || (force_vt && !COP) || CLP || !AM)
+  if (aflag || (force_vt && !D_COP) || D_CLP || !D_AM)
     {
       AddCap("xn:");
       AddCap("xv:");
       AddCap("LP:");
     }
-  if (aflag || (CS && SR) || AL || CAL)
+  if (aflag || (D_CS && D_SR) || D_AL || D_CAL)
     {
       AddCap("sr=\\EM:");
       AddCap("al=\\E[L:");
       AddCap("AL=\\E[%dL:");
     }
-  else if (SR)
+  else if (D_SR)
     AddCap("sr=\\EM:");
-  if (aflag || CS)
+  if (aflag || D_CS)
     AddCap("cs=\\E[%i%d;%dr:");
-  if (aflag || CS || DL || CDL)
+  if (aflag || D_CS || D_DL || D_CDL)
     {
       AddCap("dl=\\E[M:");
       AddCap("DL=\\E[%dM:");
     }
-  if (aflag || DC || CDC)
+  if (aflag || D_DC || D_CDC)
     {
       AddCap("dc=\\E[P:");
       AddCap("DC=\\E[%dP:");
     }
-  if (aflag || CIC || IC || IM)
+  if (aflag || D_CIC || D_IC || D_IM)
     {
       AddCap("im=\\E[4h:");
       AddCap("ei=\\E[4l:");
       AddCap("mi:");
       AddCap("IC=\\E[%d@:");
     }
+#ifdef MAPKEYS
+  AddCap("ks=\\E[?1h\\E=:");
+  AddCap("ke=\\E[?1l\\E>:");
+#endif
   if (display)
     {
-      if (US)
+      if (D_US)
 	{
 	  AddCap("us=\\E[4m:");
 	  AddCap("ue=\\E[24m:");
 	}
-      if (SO)
+      if (D_SO)
 	{
 	  AddCap("so=\\E[3m:");
 	  AddCap("se=\\E[23m:");
 	}
-      if (MB)
+      if (D_MB)
 	AddCap("mb=\\E[5m:");
-      if (MD)
+      if (D_MD)
 	AddCap("md=\\E[1m:");
-      if (MH)
+      if (D_MH)
 	AddCap("mh=\\E[2m:");
-      if (MR)
+      if (D_MR)
 	AddCap("mr=\\E[7m:");
-      if (MB || MD || MH || MR)
+      if (D_MB || D_MD || D_MH || D_MR)
 	AddCap("me=\\E[m:ms:");
-      if (VB)
+      if (D_VB)
 	AddCap("vb=\\E[?5h\\E[?5l:");
-      if (KS)
+#ifndef MAPKEYS
+      if (D_KS)
 	{
 	  AddCap("ks=\\E=:");
 	  AddCap("ke=\\E>:");
 	}
-      if (CCS)
+      if (D_CCS)
 	{
 	  AddCap("CS=\\E[?1h:");
 	  AddCap("CE=\\E[?1l:");
 	}
-      if (CG0)
+#endif
+      if (D_CG0)
 	{
 	  AddCap("G0:");
 	  AddCap("as=\\E(0:");
 	  AddCap("ae=\\E(B:");
+	  /* avoid `` because some shells dump core... */
+	  AddCap("ac=\\140\\140aaffggjjkkllmmnnooppqqrrssttuuvvwwxxyyzz{{||}}~~..--++,,hhII00:");
 	}
-      if (PO)
+      if (D_PO)
 	{
 	  AddCap("po=\\E[5i:");
 	  AddCap("pf=\\E[4i:");
 	}
-      if (CZ0)
+      if (D_CZ0)
 	{
 	  AddCap("Z0=\\E[?3h:");
 	  AddCap("Z1=\\E[?3l:");
 	}
-      if (CWS)
+      if (D_CWS)
 	AddCap("WS=\\E[8;%d;%dt:");
-      for (i = T_CAPS; i < T_ECAPS; i++)
+    }
+  for (i = T_CAPS; i < T_ECAPS; i++)
+    {
+#ifdef MAPKEYS
+      struct action *act;
+      if (i < T_OCAPS)
 	{
-	  switch(term[i].type)
+	  if (i >= T_KEYPAD)	/* don't put keypad codes in TERMCAP */
+	    continue;		/* - makes it too big */
+	  if (i >= T_CURSOR && i < T_OCAPS)
 	    {
-	    case T_STR:
-	      if (d_tcs[i].str == 0)
-		break;
-	      MakeString(term[i].tcname, buf, sizeof(buf), d_tcs[i].str);
-	      AddCap(buf);
-	      break;
-	    case T_FLG:
-	      if (d_tcs[i].flg == 0)
-		break;
-	      sprintf(buf, "%s:", term[i].tcname);
-	      AddCap(buf);
-	      break;
-	    default:
-	      break;
+	      act = &umtab[i - (T_CURSOR - T_OCAPS + T_CAPS)];
+	      if (act->nr == RC_ILLEGAL)
+		act = &dmtab[i - (T_CURSOR - T_OCAPS + T_CAPS)];
 	    }
+	  else
+	    {
+	      act = &umtab[i - T_CAPS];
+	      if (act->nr == RC_ILLEGAL)
+		act = &dmtab[i - T_CAPS];
+	    }
+	  if (act->nr != RC_ILLEGAL)
+	    {
+	      if (act->nr == RC_STUFF)
+		{
+		  MakeString(term[i].tcname, buf, sizeof(buf), act->args[0]);
+		  AddCap(buf);
+		}
+	      continue;
+	    }
+	}
+#endif
+      if (display == 0)
+	continue;
+      switch(term[i].type)
+	{
+	case T_STR:
+	  if (D_tcs[i].str == 0)
+	    break;
+	  MakeString(term[i].tcname, buf, sizeof(buf), D_tcs[i].str);
+	  AddCap(buf);
+	  break;
+	case T_FLG:
+	  if (D_tcs[i].flg == 0)
+	    break;
+	  sprintf(buf, "%s:", term[i].tcname);
+	  AddCap(buf);
+	  break;
+	default:
+	  break;
 	}
     }
   debug("MakeTermcap: end\n");
