@@ -155,7 +155,11 @@ int how;
 	}
     }
 #else	/* !UTMPOK */
-  Msg(0, "Unable to modify %s.\n", UtmpName);
+# ifdef UTMPFILE
+  Msg(0, "Unable to modify %s.\n", UTMPFILE);
+# else
+  Msg(0, "Unable to modify utmp-database.\n");
+# endif
 #endif
 }
 
